/*
 * PSP Software Development Kit - http://www.pspdev.org
 * -----------------------------------------------------------------------
 * Licensed under the BSD license, see LICENSE in PSPSDK root for details.
 *
 * main.c - Simple elf based network example.
 *
 * Copyright (c) 2005 James F <tyranid@gmail.com>
 * Some small parts (c) 2005 PSPPet
 *
 * $Id: main.c 1887 2006-05-01 03:54:06Z jim $
 * $HeadURL: svn://svn.pspdev.org/psp/trunk/pspsdk/src/samples/net/simple/main.c $
 */

#include <pspkernel.h>
#include <pspdebug.h>
#include <pspsdk.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pspnet.h>
#include <pspnet_inet.h>
#include <pspnet_apctl.h>
#include <pspnet_resolver.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/select.h>
#include <errno.h>
#include <usbPure.h>
#include <pspusb.h>
#include <pspusbstor.h>
#include <pspwlan.h>
#include <psprtc.h>

#define printf pspDebugScreenPrintf

#define MODULE_NAME "PSPRoxy"

PSP_MODULE_INFO(MODULE_NAME, 0x1000, 1, 0);
PSP_MAIN_THREAD_ATTR(0);//THREAD_ATTR_USER);

//Callbacks
int exit_callback(int arg1, int arg2, void *common)
{
	sceKernelExitGame();
	return 0;
}

int CallbackThread(SceSize args, void *argp)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);
	sceKernelSleepThreadCB();

	return 0;
}

int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread,
				     0x11, 0xFA0, PSP_THREAD_ATTR_USER, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}

//Defines
#define PACKET_SIZE 512

//Globals
unsigned char semaphore=0;
unsigned char pspRecv[PACKET_SIZE+4] __attribute__((aligned(64)));
unsigned char pspSend[PACKET_SIZE+4] __attribute__((aligned(64)));
unsigned char sockRecv[PACKET_SIZE] __attribute__((aligned(64)));
unsigned char sockSend[PACKET_SIZE] __attribute__((aligned(64)));
u64 sockSendTime=0;
u64 sockRecvTime=0;
signed int sockSendSize=0;
signed int sockRecvSize=0;
char serverIP[32];
char wifiState=0;
char semaphoreUsb=0;
char semaphoreNet=0;
char server[512];
char file[512];
char serverCurrent[512]={0};
unsigned char serverSame=0;
unsigned short serverPort=0;
int sock=-1;
struct in_addr addr;

//Functions
unsigned int resolve(char *a_server, char *a_ip)
{
  if(serverSame)
  {
    return 1;
  }
  
  if(!wifiState)
  {
    printf("Error - Wifi connection is 'no go'\n");
    return 0;
  }
  
	int rid = -1;
	char buf[256];

	/* Create a resolver */
	if(sceNetResolverCreate(&rid, buf, sizeof(buf)) < 0)
	{
		printf("Error - Could note initate resolver\n");
		if(rid >= 0)
	  {
	  	sceNetResolverDelete(rid);
	  }
    return 0;
	}

	/* Resolve a name to an ip address */
	if(sceNetResolverStartNtoA(rid, a_server, &addr, 2, 3) < 0)
	{
    printf("Error - Could not resolve %s\n", a_server);
    if(rid >= 0)
	  {
	  	sceNetResolverDelete(rid);
	  }
		return 0;
	}

	unsigned char *pointer=inet_ntoa(addr);
  int counter=0;
  while(*pointer != 0)
  {
    a_ip[counter]=*pointer;
    counter++;
    pointer++;
  }
  a_ip[counter]=0;
	
  if(rid >= 0)
	{
		sceNetResolverDelete(rid);
	}
  
  return 1;
}

unsigned int connectToServer(char *a_ip, unsigned short a_port)
{ 
  sock=socket(PF_INET, SOCK_STREAM, 0);
  if(sock < 0)
	{
    printf("Error - Could not create socket\n");
		return 0;
	}
  
  struct sockaddr_in clientService;
  clientService.sin_addr.s_addr = inet_addr(server);
  clientService.sin_family=PF_INET;
  clientService.sin_addr=addr;
  clientService.sin_port = htons(a_port);
  memset(clientService.sin_zero, 0, 8);
  
  //Connect
  if(connect(sock, &clientService, sizeof(clientService)) == -1)
  {
    printf("Error - Could not connect to server\n");
    close(sock);
    return 0;
  }
  
  return 1;
}


/* Connect to an access point */
unsigned int connectToAccessPoint(unsigned int config)
{
	int err;
	int stateLast = -1;

	/* Connect using the first profile */
	if(sceNetApctlConnect(config) != 0)
	{
		printf("Error - Could not get access to network profiles\n");
		return 0;
	}

	printf("    Connecting...\n    If it keeps displaying the same state, toggle wifi on/off\n");
	while (1)
	{
    int state;
    if(!sceWlanGetSwitchState())
    {
      printf("Error - Wireless switch is off\n");
      while(!sceWlanGetSwitchState())
      {
        sceKernelDelayThread(500000);
      }
      if(sceNetApctlConnect(config) != 0)
	    {
	    	printf("Error - Could not get access to network profiles\n");
	    	return 0;
	    }
      state=0;
      stateLast=-1;
    }
    
		
		err = sceNetApctlGetState(&state);
		if (err != 0)
		{
			printf("Error - sceNetApctlGetState returns $%x\n", err);
			break;
		}
		if (state > stateLast)
		{
      printf("    State = %d\n", state);
			stateLast = state;
		}
		if (state == 4)
			break;  // connected with static IP

		// wait a little before polling again
		sceKernelDelayThread(50*1000); // 50ms
	}
	printf("    Connected!\n");

	if(err != 0)
	{
		return 0;
	}

	return 1;
}


unsigned int net_thread(SceSize args, void *argp)
{
  struct timeval timeVal;
  signed int timeError;
  unsigned int feedBack;
  fd_set set;
	
  pspSdkInetInit();
  
  //First connect
  while(1)
	{
		if(connectToAccessPoint(1))
		{
      break;
		}
	}
  
  //Signal the USB thread to go
  feedBack=pspSdkDisableInterrupts();
  semaphore=0;
  pspSdkEnableInterrupts(feedBack);
  
  //Wait for the USB thread to finish initating
  while(!semaphore)
  {
    sceKernelDelayThread(250000);
  }
  feedBack=pspSdkDisableInterrupts();
  semaphore=0;
  pspSdkEnableInterrupts(feedBack);
  
  goto netInitiated;
  //Keep alive
  while(1)
	{
    printf("    eNet disconnected\n");
		if(connectToAccessPoint(1))
		{
      semaphoreNet=0;
    	semaphoreUsb=0;
    	
    netInitiated:
      while(1)
      {
        //Check
        wifiState=1;
        sceKernelDelayThread(2500);
        if(!sceWlanDevIsPowerOn())
        {
          sock=-1;
          semaphoreNet=0;
          semaphoreUsb=0;
          wifiState=0;
          break;
        }
        
        //Send
        if(semaphoreUsb)
        {
          if(sockSendSize <= 0)
          {
            printf("net.note=CONNECTION RESET BY CLIENT\n");
            close(sock);
            sock=-1;
            semaphoreUsb=0;           
          }
          else if(!strncmp((unsigned)sockSend, "GET http://", 11))
          {
            printf("GET detected\n");
            unsigned int counter=11;
            unsigned int counterBase=11;
            if(!strncmp((unsigned)sockSend, "GET http://www.", 15))
            {
              counter=15;
              counterBase=15;
            }
            while((sockSend[counter] != ':') && (sockSend[counter] != '/'))
            {
              server[counter-counterBase]=sockSend[counter];
              counter++;
            }
            server[counter-counterBase]=0;
            server[511]=0;
            
            serverPort=80;
            if(sockSend[counter] == ':')
            {
              //Find the port
              counter++;
              serverPort=0;
              while(sockSend[counter] != '/')
            	{
                serverPort*=10;
              	serverPort+=sockSend[counter]-'0';
              	counter++;
            	}
            }
            
            //Is it the same server?
				    serverSame=1;
            counter=0;
            while((counter < 511) && ((serverCurrent[counter] != 0) && (server[counter] != 0)))
            {
              if(serverCurrent[counter] != server[counter])
              {
                serverSame=0;
                break;
              }
              counter++;
            }
            if(counter < 4) serverSame=0;
            
            if(!serverSame)
            {
              memcpy(serverCurrent, server, 512);
            }
            
            //Generate the file link...
            counter=4;
            while((sockSend[counter] != 0x20) && (counter < 511))
            {
              file[counter-4]=sockSend[counter];
              counter++;
            }
            file[counter-4]=0;
            
            //Set the connection stats
            semaphoreNet=0;
            
            //Now connect!
            printf("GET resolving\n");
            sock=-1;
            if(resolve(server, serverIP))
            {
            	if(connectToServer(serverIP, serverPort)) printf("net.fetch=%s\n", file);
            }
            if(sock==-1)
            {
              printf("GET failed\n");
              
            }
            else
            {
              printf("GET worked\n");
           	}
          }
          
          if(sock != -1)
          {
            printf("net.send %d\n", sockSendSize);
            sockSendSize=write(sock, sockSend, sockSendSize);
            if(sockSendSize <= 0)
            {
            	printf("net.note=CONNECTION RESET BY SERVER\n");
              close(sock);
              sock=-1;
              semaphoreUsb=0;
              
              *((unsigned int*)pspSend)=0;
              semaphoreNet=0xFF;
            }
            semaphoreUsb=0;
        	}
        }
        
        //Read
        if((!semaphoreNet) && (sock != -1))
        { 
        	FD_ZERO(&set);
					FD_SET(sock, &set);
        
        	timeVal.tv_sec=1;
        	timeVal.tv_usec=0;
        	timeError=select(FD_SETSIZE, &set, NULL, NULL, &timeVal);
					if(timeError < 0)
		    	{
		    		printf("Error - Select function error\n");
		    		break;
		    	}
					
          if((timeError!=0) && FD_ISSET(sock, &set))
          {
            sockRecvSize=read(sock, sockRecv, PACKET_SIZE);
            sceRtcGetCurrentTick(&sockRecvTime);
	          if(sockRecvSize <= 0)
            {
              printf("net.note=CONNECTION RESET BY SERVER\n");
              close(sock);
              sock=-1;
              semaphoreUsb=0;
              
              *((unsigned int*)pspSend)=0;
              semaphoreNet=0xFF;
            }
            else if(sockRecvSize <= PACKET_SIZE)
            {
              printf("net.recv %d\n", sockRecvSize);
              memcpy(pspSend+4, sockRecv, PACKET_SIZE);
              
              *((unsigned int*)pspSend)=sockRecvSize;
              semaphoreNet=0xFF;
            }
          }
        }
      }
		}
	}
  
	return 0;
}

unsigned int usb_thread()
{ 
	signed int readLen;
	unsigned int feedBack;

  //Start the USB drivers
  if(sceUsbStart(PSP_USBBUS_DRIVERNAME, 0, 0) != 0) 
  {
  	printf("Error - Could not start the Main USB driver\n");
  	sceKernelSleepThread();
  }
  if(sceUsbStart(PUREUSB_DRIVER_NAME, 0, 0) != 0) 
  {
  	printf("Error - Could not start the Pure USB driver\n");
  	sceKernelSleepThread();
  }
  sceUsbActivate(PUREUSB_DRIVER_PID);
  
  //Wait for the NET thread to finish initating
  while(semaphore)
  {
    sceKernelDelayThread(250000);
  }
  
  printf("4) PSP-->iNet:\n");
  
  //Signal the NET thread to go
  feedBack=pspSdkDisableInterrupts();
  semaphore=0xFF;
  pspSdkEnableInterrupts(feedBack);
  
  //Do the loop de-loop
  goto usbInitiated;
  while(1)
  {
    printf("    iNet disconnected\n");
    semaphoreNet=0;
    semaphoreUsb=0;
    
  usbInitiated:
    printf("    Connecting...\n");
  	if(usbWaitForConnect())
    {
      printf("    Connected!\n");
    }
    else
    {
      printf("Error - Connection attempt timed out\n");
      continue;
    }

    while(1)
    {
      //Check
      sceKernelDelayThread(2500);
      if(!usbIsConnected()) break;
      
      //Send
      if(semaphoreNet)
      {
        printf("usb.Send\n");
        usbWriteData(pspSend, PACKET_SIZE+4);
        semaphoreNet=0;
      }
      
      //Read
      if(!semaphoreUsb)
      {
        readLen=usbReadData(pspRecv, PACKET_SIZE+4);
        
        //Interpret
        if(readLen == 0)
        {
					continue;
        }
        else if(readLen < 0)
        {
          printf("usb.error=WOULD YOU LIKE A BIGGER PIANO? CALL 1-800-BIG-PINO\nHomebrewers needed more money\nso they sold out to spam\nheh\n");
        	continue;
        }
        
        printf("usb.Recv->1");
        
        if(readLen==PACKET_SIZE+4)
        {
          
          printf("->2");memcpy(sockSend, pspRecv+4, PACKET_SIZE);
          printf("->3");sockSendSize=*((unsigned int*)pspRecv);
          printf("->4");sceRtcGetCurrentTick(&sockSendTime);

          printf("->5");semaphoreUsb=0xFF;
        }
        printf("\n");
      }
    }
  }
  
  return 0;
}

int LoadStartModule(char *path)
{
    u32 loadResult;
    u32 startResult;
    int status;

    loadResult = sceKernelLoadModule(path, 0, NULL);
    if (loadResult & 0x80000000)
	return -1;
    else
	startResult =
	    sceKernelStartModule(loadResult, 0, NULL, &status, NULL);

    if (loadResult != startResult)
	return -2;

    return 0;
}

/* Simple thread */
int main(int argc, char **argv)
{
  //Wait for the kernel to boot
  sceKernelDelayThread(100000);
  while(!sceKernelFindModuleByName("sceKernelLibrary"))
		sceKernelDelayThread(100000);
	sceKernelDelayThread(100000);
  
	SceUID thid;

	SetupCallbacks();
	pspDebugScreenInit();

	printf("SANiK's PSPRoxy revision A\nBuilt on TyRANiD's Simple USB driver\nhttp://sanik.imk.cx/\n\n");

	printf("1) Load USB module\n");
  if(LoadStartModule("usbPure.prx"))
  {
    printf("Error - Could not load USB module\n");
		sceKernelSleepThread();
  }

	printf("2) Load Inet module\n");
  if(pspSdkLoadInetModules() < 0)
  {
    printf("Error - Could not load Inet modules\n");
		sceKernelSleepThread();
  }
  printf("3) PSP-->eNet:\n");
  semaphore=0xFF;
	thid = sceKernelCreateThread("net_thread", net_thread, 0x18, 0x20000, PSP_THREAD_ATTR_USER, NULL);
	if(thid < 0)
	{
		printf("Error - Could not create eNet thread\n");
		sceKernelSleepThread();
	}
	sceKernelStartThread(thid, 0, NULL);

  usb_thread();
  
	//Commit suicide
	sceKernelExitDeleteThread(0);

	return 0;
}
