/*
 * PSP Software Development Kit - http://www.pspdev.org
 * -----------------------------------------------------------------------
 * Licensed under the BSD license, see LICENSE in PSPSDK root for details.
 *
 * main.c - Simple elf based network example.
 *
 * Copyright (c) 2005 James F <tyranid@gmail.com>
 * Some small parts (c) 2005 PSPPet
 *
 * $Id: main.c 1887 2006-05-01 03:54:06Z jim $
 * $HeadURL: svn://svn.pspdev.org/psp/trunk/pspsdk/src/samples/net/simple/main.c $
 */

#include <pspkernel.h>
#include <pspdebug.h>
#include <pspsdk.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pspnet.h>
#include <pspnet_inet.h>
#include <pspnet_apctl.h>
#include <pspnet_resolver.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/select.h>
#include <errno.h>
#include <usbPure.h>
#include <pspusb.h>
#include <pspusbstor.h>
#include <pspwlan.h>

#define printf pspDebugScreenPrintf

#define MODULE_NAME "PSPRoxy"

PSP_MODULE_INFO(MODULE_NAME, 0x1000, 1, 0);
PSP_MAIN_THREAD_ATTR(0);//THREAD_ATTR_USER);

//Callbacks
int exit_callback(int arg1, int arg2, void *common)
{
	sceKernelExitGame();
	return 0;
}

int CallbackThread(SceSize args, void *argp)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);
	sceKernelSleepThreadCB();

	return 0;
}

int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread,
				     0x11, 0xFA0, PSP_THREAD_ATTR_USER, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}

//Defines
#define PACKET_SIZE 512

//Globals
unsigned int bufferReadCounter=0;
unsigned int bufferWriteCounter=0;
unsigned char buffer[16][PACKET_SIZE+4];

unsigned char pspRecv[PACKET_SIZE+4] __attribute__((aligned(64)));
unsigned char pspSend[PACKET_SIZE+4] __attribute__((aligned(64)));
unsigned char sockRecv[PACKET_SIZE] __attribute__((aligned(64)));
unsigned char sockSend[PACKET_SIZE] __attribute__((aligned(64)));
signed int sockSendSize=0;
signed int sockRecvSize=0;

unsigned char serverIP[32];
unsigned short serverPort=0;
unsigned char serverCurrent[512]={0};
unsigned char serverNext[512]={0};
unsigned char serverSame=0;

unsigned char wifiState=0;
unsigned char semaphore=0;
unsigned char semaphoreUsb=0;
unsigned char semaphoreNet=0;

unsigned char file[512];


int sock=-1;
struct in_addr addr;

//Functions
unsigned int resolve(char *a_server, char *a_ip)
{
  if(serverSame)
  {
    return 1;
  }
  
  if(!wifiState)
  {
    printf("Error - Wifi connection is 'no go'\n");
    return 0;
  }
  
	int rid = -1;
	char buf[256];

	/* Create a resolver */
	if(sceNetResolverCreate(&rid, buf, sizeof(buf)) < 0)
	{
		printf("Error - Could note initate resolver\n");
		if(rid >= 0)
	  {
	  	sceNetResolverDelete(rid);
	  }
    return 0;
	}

	/* Resolve a name to an ip address */
	if(sceNetResolverStartNtoA(rid, a_server, &addr, 2, 3) < 0)
	{
    printf("Error - Could not resolve %s\n", a_server);
    if(rid >= 0)
	  {
	  	sceNetResolverDelete(rid);
	  }
		return 0;
	}

	unsigned char *pointer=inet_ntoa(addr);
  int counter=0;
  while(*pointer != 0)
  {
    a_ip[counter]=*pointer;
    counter++;
    pointer++;
  }
  a_ip[counter]=0;
	
  if(rid >= 0)
	{
		sceNetResolverDelete(rid);
	}
  
  return 1;
}

unsigned int connectToServer(char *a_ip, unsigned short a_port)
{
  sock=socket(PF_INET, SOCK_STREAM, 0);
  if(sock < 0)
	{
    printf("Error - Could not create socket\n");
		return 0;
	}
  
  struct sockaddr_in clientService;
  clientService.sin_addr.s_addr = inet_addr(serverCurrent);
  clientService.sin_family=PF_INET;
  clientService.sin_addr=addr;
  clientService.sin_port = htons(a_port);
  memset(clientService.sin_zero, 0, 8);
  
  //Connect
  if(connect(sock, &clientService, sizeof(clientService)) == -1)
  {
    printf("Error - Could not connect to server\n");
    close(sock);
    return 0;
  }
  
  return 1;
}


/* Connect to an access point */
unsigned int connectToAccessPoint(unsigned int config)
{
	int err;
	int stateLast = -1;

	/* Connect using the first profile */
	if(sceNetApctlConnect(config) != 0)
	{
		printf("Error - Could not get access to network profiles\n");
		return 0;
	}

	printf("    Connecting...\n    If it keeps displaying the same state, toggle wifi on/off\n");
	while (1)
	{
    int state;
    if(!sceWlanGetSwitchState())
    {
      printf("Error - Wireless switch is off\n");
      while(!sceWlanGetSwitchState())
      {
        sceKernelDelayThread(500000);
      }
      if(sceNetApctlConnect(config) != 0)
	    {
	    	printf("Error - Could not get access to network profiles\n");
	    	return 0;
	    }
      state=0;
      stateLast=-1;
    }
    
		
		err = sceNetApctlGetState(&state);
		if (err != 0)
		{
			printf("Error - sceNetApctlGetState returns $%x\n", err);
			break;
		}
		if (state > stateLast)
		{
      printf("    State = %d\n", state);
			stateLast = state;
		}
		if (state == 4)
			break;  // connected with static IP

		// wait a little before polling again
		sceKernelDelayThread(50*1000); // 50ms
	}
	printf("    Connected!\n");

	if(err != 0)
	{
		return 0;
	}

	return 1;
}

void buffer_next()
{
  if(bufferWriteCounter == 15)
  {
    bufferWriteCounter=0;
  }
  else
  {
    bufferWriteCounter++;
  }
}

void buffer_valid()
{
  while(*((unsigned int*)(&buffer[bufferWriteCounter][0])) != 0xFFFFFFFF)
  {
    if(semaphore) break;
    sceKernelDelayThread(2500);
  }
}

void buffer_clear()
{
  unsigned int counter=0;
  while(counter < 16)
  {
    *((unsigned int*)(&buffer[counter][0])) = 0xFFFFFFFF;
    counter++;
  }
}

void net_read()
{
  struct timeval timeVal;
  signed int timeError;
  fd_set set;
  
  do
  {
    FD_ZERO(&set);
	  FD_SET(sock, &set);
    
    timeVal.tv_sec=1;
    timeVal.tv_usec=0;
    timeError=select(FD_SETSIZE, &set, NULL, NULL, &timeVal);
	  if(timeError < 0)
	  {
	  	printf("Error - Select function error\n");
	  	break;
	  }
	
    if((timeError!=0) && FD_ISSET(sock, &set))
    {
      buffer_valid();
      
      sockRecvSize=read(sock, &buffer[bufferWriteCounter][4], PACKET_SIZE);
      sockRecvSize=sockRecvSize <= 0? 0: sockRecvSize;
      
      *((unsigned int*)(&buffer[bufferWriteCounter][0]))=sockRecvSize;
    	buffer_next();
      
      printf("net.recv %d\n", sockRecvSize);
    }
  }while(0);
}

void net_write()
{
  printf("net.send %d\n", sockSendSize);
  write(sock, sockSend, sockSendSize);
}

unsigned int net_open()
{
  if(resolve(serverCurrent, serverIP))
  {
    printf("net.fetch=%s\n", file);
  	return connectToServer(serverIP, serverPort);
  }
  else
  {
    return 0;
  }
}

void net_close()
{
  close(sock);
  sock=-1;
}

unsigned int net_thread(SceSize args, void *argp)
{
  unsigned int feedBack;
  
  pspSdkInetInit();
  
  //First connect
  while(1)
	{
		if(connectToAccessPoint(1))
		{
      break;
		}
	}
  
  //Signal the USB thread to go
  feedBack=pspSdkDisableInterrupts();
  semaphore=0;
  pspSdkEnableInterrupts(feedBack);
  
  //Wait for the USB thread to finish initating
  while(!semaphore)
  {
    sceKernelDelayThread(250000);
  }
  feedBack=pspSdkDisableInterrupts();
  semaphore=0;
  pspSdkEnableInterrupts(feedBack);
  
  goto netInitiated;
  //Keep alive
  while(1)
	{
    printf("    eNet disconnected\n");
		if(connectToAccessPoint(1))
		{
      semaphoreNet=0;
    	semaphoreUsb=0;
    	
    netInitiated:
      while(1)
      {
        //Check
        wifiState=1;
        sceKernelDelayThread(2500);
        if(!sceWlanDevIsPowerOn())
        {
          close(sock);
          sock=-1;
          semaphoreNet=0;
          semaphoreUsb=0;
          wifiState=0;
          break;
        }
        
        if(semaphoreNet) continue;
        
        //Execute the client's message
        if(semaphoreUsb)
        {
          //Connect
          if(!net_open())
          {
            semaphoreUsb=0;
            continue;
          }
          sockRecvSize=-1;
          bufferWriteCounter=0;
          bufferReadCounter=0;
          buffer_clear();
          semaphoreNet=0xFF;
          
          switch(semaphoreUsb)
          {
            case 1:	//GET file
            	net_write();
              semaphoreUsb=0;
            	while(sockRecvSize) {if(semaphore){break;} net_read();}
            	break;
            case 0x80:
            	net_write();
              semaphoreUsb=0;
            	break;
          }
          
          //Disconnect
          net_close();
          semaphoreNet=0x80;
        }
      }
		}
	}
  
	return 0;
}

unsigned int usb_thread()
{ 
	signed int readLen;
	unsigned int feedBack;

  //Start the USB drivers
  if(sceUsbStart(PSP_USBBUS_DRIVERNAME, 0, 0) != 0) 
  {
  	printf("Error - Could not start the Main USB driver\n");
  	sceKernelSleepThread();
  }
  if(sceUsbStart(PUREUSB_DRIVER_NAME, 0, 0) != 0) 
  {
  	printf("Error - Could not start the Pure USB driver\n");
  	sceKernelSleepThread();
  }
  sceUsbActivate(PUREUSB_DRIVER_PID);
  
  //Wait for the NET thread to finish initating
  while(semaphore)
  {
    sceKernelDelayThread(250000);
  }
  
  printf("4) PSP-->iNet:\n");
  
  //Signal the NET thread to go
  feedBack=pspSdkDisableInterrupts();
  semaphore=0xFF;
  pspSdkEnableInterrupts(feedBack);
  
  //Do the loop de-loop
  goto usbInitiated;
  while(1)
  {
    printf("    iNet disconnected\n");
    semaphoreNet=0;
    semaphoreUsb=0;
    
  usbInitiated:
    printf("    Connecting...\n");
  	if(usbWaitForConnect())
    {
      printf("    Connected!\n");
    }
    else
    {
      printf("Error - Connection attempt timed out\n");
      continue;
    }

    while(1)
    {
      //Check
      sceKernelDelayThread(25000);
      if(!usbIsConnected()) break;
      
      //Send
      if(semaphoreNet)
      {	
      	if(semaphore)
        {
          while(semaphoreNet == 0x80)
          {
            sceKernelDelayThread(2500);
          }
          semaphore=0;
          semaphoreNet=0;
          bufferReadCounter=0;
          continue;
        }
        if(*((unsigned int*)(&buffer[bufferReadCounter][0])) != 0xFFFFFFFF)
        {
          	printf("usb.send %d\n", *((unsigned int*)(&buffer[bufferReadCounter][0])));
        		usbWriteData(&buffer[bufferReadCounter][0], PACKET_SIZE+4);
          	*((unsigned int*)(&buffer[bufferReadCounter][0]))=0xFFFFFFFF;
          
          bufferReadCounter++;
          if(bufferReadCounter == 16)
          {
            bufferReadCounter=0;
          }
          if(*((unsigned int*)(&buffer[bufferReadCounter][0])) != 0xFFFFFFFF)
          {
            continue;
          }
        }
        
        if(semaphoreNet == 0x80)	//Was that the last transfer?
        {
          semaphoreNet=0;
          bufferReadCounter=0;
        }
      }
      
      //Read
      if(!semaphoreUsb)
      {
        readLen=usbReadData(pspRecv, PACKET_SIZE+4);
        
        if(readLen == 0)
        {
          //Timed out
					continue;
        }
        else if(readLen < 0)
        {
          printf("usb.error=WOULD YOU LIKE A BIGGER PIANO? CALL 1-800-BIG-PINO\nHomebrewers needed more money\nso they sold out to spam\nheh\n");
        	continue;
        }
        
        //Interpret
        memcpy(sockSend, pspRecv+4, PACKET_SIZE);
        sockSendSize=*((unsigned int*)pspRecv);
        printf("usb.recv %d=", sockSendSize);
        
        if(readLen==PACKET_SIZE+4)
        {
          if(sockSendSize == 0)
          {
            //Stop the transfer
            semaphore=1;
            printf("\n");
          }
          else
          {  
            int counter=0;
            while(counter < 50)
            {
              printf("%c", sockSend[counter] <= 0x20? '.': sockSend[counter]);
              counter++;
            }
            printf("\n");
            
            if(!strncmp((unsigned)sockSend, "GET http://", 11))	//Get page?
            {
              //Get the server name
              unsigned int counter=11;
              unsigned int counterBase=11;
              if(!strncmp((unsigned)sockSend, "GET http://www.", 15))
              {
                counter=15;
                counterBase=15;
              }
              while(((counter-counterBase) < 511) && (sockSend[counter] != ':') && (sockSend[counter] != '/'))
              {
                serverNext[counter-counterBase]=sockSend[counter];
                counter++;
              }
              serverNext[counter-counterBase]=0;
              
              //Get the port number
              if(sockSend[counter] == ':')
              {
                //Find the port
                counter++;
                serverPort=0;
                while(sockSend[counter] != '/')
              	{
                  serverPort*=10;
                	serverPort+=sockSend[counter]-'0';
                	counter++;
              	}
              }
              else
              {
                serverPort=80;
              }
              
              //Is it the same server?
				      serverSame=1;
              counter=0;
              while((counter < 511) && ((serverCurrent[counter] != 0) && (serverNext[counter] != 0)))
              {
                if(serverCurrent[counter] != serverNext[counter])
                {
                  serverSame=0;
                  break;
                }
                counter++;
              }
              if(counter < 4) serverSame=0;
              
              if(!serverSame)
              {
                memcpy(serverCurrent, serverNext, 512);
              }
              
              //Generate the file link...
              counter=4;
              while((sockSend[counter] != 0x20) && (counter < 511))
              {
                file[counter-4]=sockSend[counter];
                counter++;
              }
              file[counter-4]=0;
              
              //Set the connection signal
              semaphoreUsb=1;	//Get page
            }
            else
            {
              printf("Unknown HTTP command\n");
              semaphoreUsb=0x80; //Transfer like normal
            }
            /*else if(!strncmp((unsigned)sockSend, "GET http://", 11))
            {
              
            }*/
          }
        }
      }
      else
      {
        printf("Can't sir =/ %d\n", semaphoreUsb);
      }
    }
  }
  
  return 0;
}

int LoadStartModule(char *path)
{
    u32 loadResult;
    u32 startResult;
    int status;

    loadResult = sceKernelLoadModule(path, 0, NULL);
    if (loadResult & 0x80000000)
	return -1;
    else
	startResult =
	    sceKernelStartModule(loadResult, 0, NULL, &status, NULL);

    if (loadResult != startResult)
	return -2;

    return 0;
}

/* Simple thread */
int main(int argc, char **argv)
{
  //Wait for the kernel to boot
  sceKernelDelayThread(100000);
  while(!sceKernelFindModuleByName("sceKernelLibrary"))
		sceKernelDelayThread(100000);
	sceKernelDelayThread(100000);
  
	SceUID thid;

	SetupCallbacks();
	pspDebugScreenInit();

	printf("SANiK's PSPRoxy revision A\nBuilt on TyRANiD's Simple USB driver\nhttp://sanik.imk.cx/\n\n");

	printf("1) Load USB module\n");
  if(LoadStartModule("usbPure.prx"))
  {
    printf("Error - Could not load USB module\n");
		sceKernelSleepThread();
  }

	printf("2) Load Inet module\n");
  if(pspSdkLoadInetModules() < 0)
  {
    printf("Error - Could not load Inet modules\n");
		sceKernelSleepThread();
  }
  printf("3) PSP-->eNet:\n");
  semaphore=0xFF;
	thid = sceKernelCreateThread("net_thread", net_thread, 0x18, 0x20000, PSP_THREAD_ATTR_USER, NULL);
	if(thid < 0)
	{
		printf("Error - Could not create eNet thread\n");
		sceKernelSleepThread();
	}
	sceKernelStartThread(thid, 0, NULL);

  usb_thread();
  
	//Commit suicide
	sceKernelExitDeleteThread(0);

	return 0;
}
