/*
 * PSPLINK
 * -----------------------------------------------------------------------
 * Licensed under the BSD license, see LICENSE in PSPLINK root for details.
 *
 * usbasync.h - PSPLINK USB Asynchronous Data Provider
 *
 * Copyright (c) 2006 James F
 *
 * $HeadURL: file:///home/svn/psp/branches/psplinkusb/usbhostfs/usbPure.h $
 * $Id: usbasync.h 2177 2007-02-15 17:28:02Z tyranid $
 */
#ifndef __USBASYNC_H__
#define __USBASYNC_H__

#include <stdint.h>

#define PUREUSB_MAX_BLOCK (64*1024)
#define PUREUSB_BULK_MAXWRITE  (1024*1024)
#define PUREUSB_MODULE_NAME "USBPure"
#define PUREUSB_DRIVER_NAME "USBPure"
#define PUREUSB_DRIVER_PID  (0x1C9)
#define PUREUSB_SONY_VID (0x54C)
#define PUREUSB_MAGIC  0x08681961   

/**
  * Wait for the USB connection to be established
  *
  * @return 0 on connected, < 0 on error
  */
int usbWaitForConnect(void);

/**
  * Return the status of the USB connection
  *
  * @return 0 on connected, < 0 on error
  */
int usbIsConnected(void);

/**
 * Write a data transfer block
 *
 * @param data - Pointer to the data to write (should be 16byte aligned)
 * @param size - Size of data set to write
 *
 * @return the number of bytes written, < 0 on error
 */
int usbWriteData(const void *data, int size);

/**
 * Read a data transfer block
 *
 * @param data - Pointer to the data to write (should be 16byte aligned)
 * @param size - Size of data set to write
 *
 * @return the number of bytes written, < 0 on error
 */
int usbReadData(void *data, int size);

/**
 * Lock the USB bus (normally to ensure no threads are using it)
 *
 * @return < 0 on error
 */
int usbLockBus(void);

/**
 * Unlock the USB bus 
 *
 * @return < 0 on error
 */
int usbUnlockBus(void);

#endif
